import ujson as json
import unittest

from requests.exceptions import ConnectionError

from betfairlightweight import APIClient
from betfairlightweight.endpoints.baseendpoint import BaseEndpoint
from betfairlightweight.exceptions import (
    APIError,
    InvalidResponse,
)
from tests import mock
from tests.unit.tools import create_mock_json


class BaseEndpointInit(unittest.TestCase):

    def test_base_endpoint_init(self):
        client = APIClient('username', 'password', 'app_key')
        base_endpoint = BaseEndpoint(client)
        assert base_endpoint.connect_timeout == 3.05
        assert base_endpoint.read_timeout == 16
        assert base_endpoint._error == APIError
        assert base_endpoint.client == client


class BaseEndPointTest(unittest.TestCase):

    def setUp(self):
        client = APIClient('username', 'password', 'app_key', 'UK')
        self.base_endpoint = BaseEndpoint(client)

    def test_base_endpoint_create_req(self):
        payload = {'jsonrpc': '2.0',
                   'method': 'test',
                   'params': 'empty',
                   'id': 1}
        assert self.base_endpoint.create_req('test', 'empty') == json.dumps(payload)

    @mock.patch('betfairlightweight.endpoints.baseendpoint.BaseEndpoint.create_req')
    @mock.patch('betfairlightweight.baseclient.BaseClient.cert')
    @mock.patch('betfairlightweight.baseclient.BaseClient.request_headers')
    @mock.patch('betfairlightweight.baseclient.requests.post')
    def test_request(self, mock_post, mock_request_headers, mock_cert, mock_create_req):
        mock_response = create_mock_json('tests/resources/login_success.json')
        mock_post.return_value = mock_response

        mock_client_cert = mock.Mock()
        mock_client_cert.return_value = []
        mock_cert.return_value = mock_client_cert

        url = 'https://api.betfair.com/exchange/betting/json-rpc/v1'
        response = self.base_endpoint.request(None, None, None)

        mock_post.assert_called_once_with(url, data=mock_create_req(),
                                          headers=mock_request_headers, timeout=(3.05, 16))
        assert response[0] == mock_response.json()
        assert isinstance(response[1], float)

    @mock.patch('betfairlightweight.endpoints.baseendpoint.BaseEndpoint.create_req')
    @mock.patch('betfairlightweight.baseclient.BaseClient.cert')
    @mock.patch('betfairlightweight.baseclient.BaseClient.request_headers')
    @mock.patch('betfairlightweight.baseclient.requests.post')
    def test_request_error(self, mock_post, mock_request_headers, mock_cert, mock_create_req):
        mock_post.side_effect = ConnectionError()
        with self.assertRaises(APIError):
            self.base_endpoint.request(None, None, None)

        mock_post.side_effect = ValueError()
        with self.assertRaises(APIError):
            self.base_endpoint.request(None, None, None)

    @mock.patch('betfairlightweight.endpoints.baseendpoint.BaseEndpoint.create_req')
    @mock.patch('betfairlightweight.baseclient.BaseClient.cert')
    @mock.patch('betfairlightweight.baseclient.BaseClient.request_headers')
    @mock.patch('betfairlightweight.baseclient.requests.post')
    def test_request_json_error(self, mock_post, mock_request_headers, mock_cert, mock_create_req):
        mock_response = mock.Mock()
        mock_response.status_code = 200
        mock_response.json.side_effect = ValueError()
        mock_post.return_value = mock_response

        with self.assertRaises(InvalidResponse):
            self.base_endpoint.request(None, None, None)

    def test_base_endpoint_error_handler(self):
        mock_response = create_mock_json('tests/resources/base_endpoint_success.json')
        assert self.base_endpoint._error_handler(mock_response.json()) is None

        mock_response = create_mock_json('tests/resources/base_endpoint_fail.json')
        with self.assertRaises(APIError):
            self.base_endpoint._error_handler(mock_response.json())

    def test_base_endpoint_process_response(self):
        mock_resource = mock.Mock()

        response_list = [{}, {}]
        response = self.base_endpoint.process_response(response_list, mock_resource, None, False)
        assert type(response) == list
        assert response[0] == mock_resource()

        response_result_list = {'result': [{}, {}]}
        response = self.base_endpoint.process_response(response_result_list, mock_resource, None, False)
        assert type(response) == list
        assert response[0] == mock_resource()

        response_result = {'result': {}}
        response = self.base_endpoint.process_response(response_result, mock_resource, None, False)
        assert response == mock_resource()

        # lightweight tests
        response_list = [{}, {}]
        response = self.base_endpoint.process_response(response_list, mock_resource, None, True)
        assert response == response_list

        client = APIClient('username', 'password', 'app_key', lightweight=True)
        base_endpoint = BaseEndpoint(client)
        response_list = [{}, {}]
        response = base_endpoint.process_response(response_list, mock_resource, None, False)
        assert type(response) == list
        assert response[0] == mock_resource()

    def test_base_endpoint_process_response_no_error(self):

        class MockResource(object):
            def __init__(self, elapsed_time, hello):
                self.elapsed_time = elapsed_time
                self.hello = hello

        mock_resource = MockResource
        response_list = [{'hello': 'world'}]

        response = self.base_endpoint.process_response(response_list, mock_resource, 0, False)
        assert type(response) == list
        assert response[0].__dict__ == mock_resource(elapsed_time=0, **response_list[0]).__dict__

    def test_base_endpoint_process_response_error(self):

        class MockResource(object):
            def __init__(self, elapsed_time, hello):
                self.elapsed_time = elapsed_time
                self.hello = hello

        mock_resource = MockResource
        response_list = [{}]  # missing 'hello'

        with self.assertRaises(InvalidResponse):
            self.base_endpoint.process_response(response_list, mock_resource, 0, False)

    def test_base_endpoint_url(self):
        assert self.base_endpoint.url == '%s%s' % (self.base_endpoint.client.api_uri, 'betting/json-rpc/v1')





class BetfairError(Exception):
    """
    Base class for Betfair Errors.
    """
    pass


class PasswordError(BetfairError):
    """
    Exception raised if password is not found.
    """

    def __init__(self, username):
        message = 'Password not found in .bashprofile for %s, add or pass to APIClient' % username
        super(PasswordError, self).__init__(message)


class AppKeyError(BetfairError):
    """
    Exception raised if appkey is not found.
    """

    def __init__(self, username):
        message = 'AppKey not found in .bashprofile for %s, add or pass to APIClient' % username
        super(AppKeyError, self).__init__(message)


class CertsError(BetfairError):
    """
    Exception raised if certs folder is not found.
    """

    def __init__(self, path='/certs/'):
        message = 'Certificate folder not found in %s' % path
        super(CertsError, self).__init__(message)


class StatusCodeError(BetfairError):
    """
    Exception raised if status code is incorrect.
    """

    def __init__(self, status_code):
        message = 'Status code error: %s' % status_code
        super(StatusCodeError, self).__init__(message)


class InvalidResponse(BetfairError):
    """
    Exception raised if invalid response
    received from betfair.
    """

    def __init__(self, response):
        message = 'Invalid response received: %s' % response
        super(InvalidResponse, self).__init__(message)


class LoginError(BetfairError):
    """
    Exception raised if sessionToken is not found.
    """

    def __init__(self, response):
        login_status = response.get('loginStatus')
        if login_status is None:  # different response when interactive login requested
            login_status = response.get('error', 'UNKNOWN')
        message = 'API login: %s' % login_status
        super(LoginError, self).__init__(message)


class KeepAliveError(BetfairError):
    """
    Exception raised if keep alive fails.
    """

    def __init__(self, response):
        keep_alive_status = response.get('status', 'UNKNOWN')
        keep_alive_error = response.get('error')
        message = 'API keepAlive %s: %s' % (keep_alive_status, keep_alive_error)
        super(KeepAliveError, self).__init__(message)


class APIError(BetfairError):
    """
    Exception raised if error is found.
    """

    def __init__(self, response, method=None, params=None, exception=None):
        if response:
            error_data = response.get('error')
            message = '%s \nParams: %s \nException: %s \nError: %s \nFull Response: %s' % (
                method, params, exception, error_data, response
            )
        else:
            message = '%s \nParams: %s \nException: %s' % (
                method, params, exception
            )
        super(APIError, self).__init__(message)


class LogoutError(BetfairError):
    """
    Exception raised if logout errors.
    """

    def __init__(self, response):
        logout_status = response.get('status', 'UNKNOWN')
        logout_error = response.get('error')
        message = 'API logout %s: %s' % (logout_status, logout_error)
        super(LogoutError, self).__init__(message)


# class ParameterError(BetfairError):
#     """Exception raised if parameter is incorrect"""
#
#     def __init__(self, api_method):
#         message = 'API method %s must have parameters' % api_method
#         super(ParameterError, self).__init__(message)
#
#
# class SessionTokenError(BetfairError):
#     """Exception raised if session_token is None"""
#
#     def __init__(self):
#         message = 'APIClient must have session_token'
#         super(SessionTokenError, self).__init__(message)


class SocketError(BetfairError):
    """
    Exception raised if error with socket.
    """

    def __init__(self, message):
        super(SocketError, self).__init__(message)


class ListenerError(BetfairError):
    """
    Exception raised if error with listener.
    """

    def __init__(self, connection_id, data):
        message = 'connection_id: %s, data: %s' % (connection_id, data)
        super(ListenerError, self).__init__(message)


class CacheError(BetfairError):
    """
    Exception raised if error with cache.
    """

    def __init__(self, message):
        super(CacheError, self).__init__(message)


class RaceCardError(BetfairError):
    """
    Exception raised if error with race card request.
    """

    def __init__(self, message):
        super(RaceCardError, self).__init__(message)



import unittest

from betfairlightweight.resources.streamingresources import (
    MarketDefinition,
    MarketDefinitionRunner,
)
from tests.unit.tools import create_mock_json


class TestMarketDefinition(unittest.TestCase):

    def setUp(self):
        self.mock_response = create_mock_json('tests/resources/streaming_market_definition.json')
        self.market_definition = MarketDefinition(**self.mock_response.json())

    def test_init(self):
        assert len(self.market_definition.runners) == 7
        assert self.market_definition.bsp_market is True
        assert self.market_definition.market_base_rate == 5
        assert len(self.market_definition.key_line_definitions.key_line) == 2

    def test_missing_open_date(self):
        response_json = dict(self.mock_response.json())
        response_json.pop('openDate')
        market_definition = MarketDefinition(**response_json)
        assert market_definition.open_date is None


class TestMarketDefinitionRunner(unittest.TestCase):

    def setUp(self):
        self.mock_response = create_mock_json('tests/resources/streaming_market_definition.json')
        market_definition = self.mock_response.json()
        runner = market_definition['runners'][0]
        self.market_definition_runner = MarketDefinitionRunner(**runner)

    def test_init(self):
        assert self.market_definition_runner.selection_id == 11131804
        assert self.market_definition_runner.adjustment_factor == 44.323
        assert self.market_definition_runner.sort_priority == 1
        assert self.market_definition_runner.status == 'ACTIVE'
        assert self.market_definition_runner.removal_date is None

    def test_str(self):
        assert str(self.market_definition_runner) == 'MarketDefinitionRunner: 11131804'

    def test_repr(self):
        assert repr(self.market_definition_runner) == '<MarketDefinitionRunner>'



from .baseresource import BaseResource


class AccountFunds(BaseResource):

    def __init__(self, **kwargs):
        super(AccountFunds, self).__init__(**kwargs)
        self.available_to_bet_balance = kwargs.get('availableToBetBalance')
        self.discount_rate = kwargs.get('discountRate')
        self.exposure = kwargs.get('exposure')
        self.exposure_limit = kwargs.get('exposureLimit')
        self.points_balance = kwargs.get('pointsBalance')
        self.retained_commission = kwargs.get('retainedCommission')
        self.wallet = kwargs.get('wallet')


class AccountDetails(BaseResource):

    def __init__(self, **kwargs):
        super(AccountDetails, self).__init__(**kwargs)
        self.country_code = kwargs.get('countryCode')
        self.currency_code = kwargs.get('currencyCode')
        self.discount_rate = kwargs.get('discountRate')
        self.first_name = kwargs.get('firstName')
        self.last_name = kwargs.get('lastName')
        self.locale_code = kwargs.get('localeCode')
        self.points_balance = kwargs.get('pointsBalance')
        self.region = kwargs.get('region')
        self.timezone = kwargs.get('timezone')


class LegacyData(object):

    def __init__(self, avgPrice, betCategoryType, betSize, betType, eventId, eventTypeId, fullMarketName, marketName,
                 grossBetAmount, transactionId, marketType, placedDate, selectionId, startDate, transactionType,
                 winLose, selectionName=None, commissionRate=None):
        self.avg_price = avgPrice
        self.bet_category_type = betCategoryType
        self.bet_size = betSize
        self.bet_type = betType
        self.event_id = eventId
        self.event_type_id = eventTypeId
        self.full_market_name = fullMarketName
        self.gross_bet_amount = grossBetAmount
        self.market_name = marketName
        self.transaction_id = transactionId
        self.market_type = marketType
        self.placed_date = BaseResource.strip_datetime(placedDate)
        self.selection_id = selectionId
        self.start_date = BaseResource.strip_datetime(startDate)
        self.transaction_type = transactionType
        self.win_lose = winLose
        self.selection_name = selectionName
        self.commission_rate = commissionRate


class AccountStatement(object):

    def __init__(self, amount, balance, itemClass, itemClassData, itemDate, refId, legacyData):
        self.amount = amount
        self.balance = balance
        self.item_class = itemClass
        self.item_class_data = itemClassData
        self.item_date = BaseResource.strip_datetime(itemDate)
        self.ref_id = refId
        self.legacy_data = LegacyData(**legacyData)


class AccountStatementResult(BaseResource):

    def __init__(self, **kwargs):
        super(AccountStatementResult, self).__init__(**kwargs)
        self.more_available = kwargs.get('moreAvailable')
        self.account_statement = [AccountStatement(**i) for i in kwargs.get('accountStatement')]


class CurrencyRate(BaseResource):

    def __init__(self, **kwargs):
        super(CurrencyRate, self).__init__(**kwargs)
        self.currency_code = kwargs.get('currencyCode')
        self.rate = kwargs.get('rate')


class TransferFunds(BaseResource):

    def __init__(self, **kwargs):
        super(TransferFunds, self).__init__(**kwargs)
        self.transaction_id = kwargs.get('transactionId')



from __future__ import print_function

import datetime
import unittest

from betfairlightweight import resources
from betfairlightweight.resources.bettingresources import (
    LimitOrder,
    LimitOnCloseOrder,
    MarketOnCloseOrder,
    PriceSize,
)
from tests.unit.tools import create_mock_json


class BettingResourcesTest(unittest.TestCase):

    ELAPSED_TIME = 1.2

    def test_event_type_result(self):
        mock_response = create_mock_json('tests/resources/list_event_types.json')
        event_types = mock_response.json().get('result')

        for event_type in event_types:
            resource = resources.EventTypeResult(elapsed_time=self.ELAPSED_TIME,
                                                 **event_type)

            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_count == event_type['marketCount']
            assert resource.event_type.id == event_type['eventType']['id']
            assert resource.event_type.name == event_type['eventType']['name']

    def test_competition_result(self):
        mock_response = create_mock_json('tests/resources/list_competitions.json')
        competitions = mock_response.json().get('result')

        for competition in competitions:
            resource = resources.CompetitionResult(elapsed_time=self.ELAPSED_TIME,
                                                   **competition)

            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_count == competition['marketCount']
            assert resource.competition_region == competition['competitionRegion']
            assert resource.competition.id == competition['competition']['id']
            assert resource.competition.name == competition['competition']['name']

    def test_time_range_result(self):
        mock_response = create_mock_json('tests/resources/list_time_ranges.json')
        time_ranges = mock_response.json().get('result')

        for time_range in time_ranges:
            resource = resources.TimeRangeResult(elapsed_time=self.ELAPSED_TIME,
                                                 **time_range)

            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_count == time_range['marketCount']
            assert resource.time_range._from == datetime.datetime.strptime(time_range['timeRange']['from'],
                                                                           "%Y-%m-%dT%H:%M:%S.%fZ")
            assert resource.time_range.to == datetime.datetime.strptime(time_range['timeRange']['to'],
                                                                        "%Y-%m-%dT%H:%M:%S.%fZ")

    def test_event_result(self):
        mock_response = create_mock_json('tests/resources/list_events.json')
        event_results = mock_response.json().get('result')

        for event_result in event_results:
            resource = resources.EventResult(elapsed_time=self.ELAPSED_TIME,
                                             **event_result)

            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_count == event_result['marketCount']
            assert resource.event.id == event_result['event']['id']
            assert resource.event.open_date == datetime.datetime.strptime(event_result['event']['openDate'],
                                                                          "%Y-%m-%dT%H:%M:%S.%fZ")
            assert resource.event.time_zone == event_result['event']['timezone']
            assert resource.event.country_code == event_result['event']['countryCode']
            assert resource.event.name == event_result['event']['name']
            assert resource.event.venue == event_result['event']['venue']

    def test_market_type_result(self):
        mock_response = create_mock_json('tests/resources/list_market_types.json')
        market_type_results = mock_response.json().get('result')

        for market_type_result in market_type_results:
            resource = resources.MarketTypeResult(elapsed_time=self.ELAPSED_TIME,
                                                  **market_type_result)

            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_count == market_type_result['marketCount']
            assert resource.market_type == market_type_result['marketType']

    def test_country_result(self):
        mock_response = create_mock_json('tests/resources/list_countries.json')
        countries_results = mock_response.json().get('result')

        for countries_result in countries_results:
            resource = resources.CountryResult(elapsed_time=self.ELAPSED_TIME,
                                               **countries_result)

            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_count == countries_result['marketCount']
            assert resource.country_code == countries_result['countryCode']

    def test_venue_result(self):
        mock_response = create_mock_json('tests/resources/list_venues.json')
        venue_results = mock_response.json().get('result')

        for venue_result in venue_results:
            resource = resources.VenueResult(elapsed_time=self.ELAPSED_TIME,
                                             **venue_result)

            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_count == venue_result['marketCount']
            assert resource.venue == venue_result['venue']

    def test_market_catalogue(self):
        mock_response = create_mock_json('tests/resources/list_market_catalogue.json')
        market_catalogues = mock_response.json().get('result')

        for market_catalogue in market_catalogues:
            resource = resources.MarketCatalogue(elapsed_time=self.ELAPSED_TIME,
                                                 **market_catalogue)

            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_id == market_catalogue['marketId']
            assert resource.market_name == market_catalogue['marketName']
            assert resource.total_matched == market_catalogue['totalMatched']
            assert resource.market_start_time == datetime.datetime.strptime(
                    market_catalogue['marketStartTime'], "%Y-%m-%dT%H:%M:%S.%fZ")

            assert resource.competition.id == market_catalogue['competition']['id']
            assert resource.competition.name == market_catalogue['competition']['name']

            assert resource.event.id == market_catalogue['event']['id']
            assert resource.event.open_date == datetime.datetime.strptime(market_catalogue['event']['openDate'],
                                                                          "%Y-%m-%dT%H:%M:%S.%fZ")
            assert resource.event.time_zone == market_catalogue['event']['timezone']
            assert resource.event.country_code == market_catalogue['event']['countryCode']
            assert resource.event.name == market_catalogue['event']['name']
            assert resource.event.venue == market_catalogue['event'].get('venue')

            assert resource.event_type.id == market_catalogue['eventType']['id']
            assert resource.event_type.name == market_catalogue['eventType']['name']

            assert resource.description.betting_type == market_catalogue['description']['bettingType']
            assert resource.description.bsp_market == market_catalogue['description']['bspMarket']
            assert resource.description.discount_allowed == market_catalogue['description']['discountAllowed']
            assert resource.description.market_base_rate == market_catalogue['description']['marketBaseRate']
            assert resource.description.market_time == datetime.datetime.strptime(
                    market_catalogue['description']['marketTime'], "%Y-%m-%dT%H:%M:%S.%fZ")
            assert resource.description.market_type == market_catalogue['description']['marketType']
            assert resource.description.persistence_enabled == market_catalogue['description']['persistenceEnabled']
            assert resource.description.regulator == market_catalogue['description']['regulator']
            assert resource.description.rules == market_catalogue['description']['rules']
            assert resource.description.rules_has_date == market_catalogue['description']['rulesHasDate']
            assert resource.description.suspend_time == datetime.datetime.strptime(
                    market_catalogue['description']['suspendTime'], "%Y-%m-%dT%H:%M:%S.%fZ")
            assert resource.description.turn_in_play_enabled == market_catalogue['description']['turnInPlayEnabled']
            assert resource.description.wallet == market_catalogue['description']['wallet']
            assert resource.description.each_way_divisor == market_catalogue['description'].get('eachWayDivisor')
            assert resource.description.clarifications == market_catalogue['description'].get('clarifications')
            assert resource.description.line_range_info.interval  == market_catalogue['description']['lineRangeInfo'].get('interval')
            assert resource.description.line_range_info.market_unit == market_catalogue['description']['lineRangeInfo'].get('marketUnit')
            assert resource.description.line_range_info.min_unit_value  == market_catalogue['description']['lineRangeInfo'].get('minUnitValue')
            assert resource.description.line_range_info.max_unit_value  == market_catalogue['description']['lineRangeInfo'].get('maxUnitValue')
            assert resource.description.price_ladder_description.type == market_catalogue['description']['priceLadderDescription'].get('type')

            assert len(resource.runners) == 10
            assert resource.runners[6].handicap == 0.0
            assert resource.runners[6].runner_name == 'SCR Altach'
            assert resource.runners[6].selection_id == 872710
            assert resource.runners[6].sort_priority == 7
            assert resource.runners[6].metadata == {'runnerId': '872710'}

    def test_market_book(self):
        mock_response = create_mock_json('tests/resources/list_market_book.json')
        market_books = mock_response.json().get('result')

        for market_book in market_books:
            resource = resources.MarketBook(elapsed_time=self.ELAPSED_TIME,
                                            **market_book)
            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_id == market_book['marketId']
            assert resource.bet_delay == market_book['betDelay']
            assert resource.bsp_reconciled == market_book['bspReconciled']
            assert resource.complete == market_book['complete']
            assert resource.cross_matching == market_book['crossMatching']
            assert resource.inplay == market_book['inplay']
            assert resource.is_market_data_delayed == market_book['isMarketDataDelayed']
            assert resource.last_match_time == datetime.datetime.strptime(
                    market_book['lastMatchTime'], "%Y-%m-%dT%H:%M:%S.%fZ")
            assert resource.number_of_active_runners == market_book['numberOfActiveRunners']
            assert resource.number_of_runners == market_book['numberOfRunners']
            assert resource.number_of_winners == market_book['numberOfWinners']
            assert resource.runners_voidable == market_book['runnersVoidable']
            assert resource.status == market_book['status']
            assert resource.total_available == market_book['totalAvailable']
            assert resource.total_matched == market_book['totalMatched']
            assert resource.version == market_book['version']

            assert len(resource.runners) == len(market_book['runners'])

            for i, key_line in enumerate(market_book['keyLineDescription'].get('keyLine', [])):
                assert key_line['handicap'] == resource.key_line_description.key_line[i].handicap
                assert key_line['selectionId'] == resource.key_line_description.key_line[i].selection_id

            for i, runner in enumerate(market_book['runners']):
                resource_runner = resource.runners[i]
                assert resource_runner.selection_id == runner['selectionId']
                assert resource_runner.status == runner['status']
                assert resource_runner.total_matched == runner.get('totalMatched')
                assert resource_runner.adjustment_factor == runner.get('adjustmentFactor')
                assert resource_runner.handicap == runner['handicap']
                assert resource_runner.last_price_traded == runner.get('lastPriceTraded')

                if runner.get('removalDate'):
                    assert resource_runner.removal_date == datetime.datetime.strptime(
                        runner['removalDate'], "%Y-%m-%dT%H:%M:%S.%fZ")
                # else:
                #     assert resource_runner.sp.near_price == runner['sp']['nearPrice']
                #     assert resource_runner.sp.far_price == runner['sp']['farPrice']
                #     assert resource_runner.sp.actual_sp == runner['sp']['actualSP']

                # assert resource_runner.sp.back_stake_taken == runner['sp']['backStakeTaken']
                # assert resource_runner.sp.lay_liability_taken == runner['sp']['layLiabilityTaken']
                #
                # assert resource_runner.ex.available_to_back == runner['ex'].get('availableToBack')
                # assert resource_runner.ex.available_to_lay == runner['ex'].get('availableToLay')
                # assert resource_runner.ex.traded_volume == runner['ex'].get('tradedVolume')

                # # print(resource_runner.orders)
                # # print(resource_runner.matches)
                # # todo complete

    def test_price_size(self):
        price_size = PriceSize(**{'price': 1.01, 'size': 2048})
        self.assertEqual(price_size.price, 1.01)
        self.assertEqual(price_size.size, 2048)
        self.assertEqual(str(price_size), 'Price: 1.01 Size: 2048')

    def test_current_orders(self):
        mock_response = create_mock_json('tests/resources/list_current_orders.json')
        current_orders = mock_response.json().get('result')
        resource = resources.CurrentOrders(elapsed_time=self.ELAPSED_TIME,
                                           **current_orders)
        assert resource.elapsed_time == self.ELAPSED_TIME
        assert len(resource.orders) == len(current_orders.get('currentOrders'))

        for current_order in current_orders.get('currentOrders'):
            assert resource.orders[0].bet_id == current_order['betId']
            # todo complete

    def test_cleared_orders(self):
        mock_response = create_mock_json('tests/resources/list_cleared_orders.json')
        cleared_orders = mock_response.json().get('result')
        resource = resources.ClearedOrders(elapsed_time=self.ELAPSED_TIME,
                                           **cleared_orders)
        assert resource.elapsed_time == self.ELAPSED_TIME
        assert len(resource.orders) == len(cleared_orders.get('clearedOrders'))

        for cleared_order in cleared_orders.get('clearedOrders'):
            assert resource.orders[0].bet_id == cleared_order['betId']
            # todo complete

    def test_market_profit_loss(self):
        mock_response = create_mock_json('tests/resources/list_market_profit_and_loss.json')
        market_profits = mock_response.json().get('result')

        for market_profit in market_profits:
            resource = resources.MarketProfitLoss(elapsed_time=self.ELAPSED_TIME,
                                                  **market_profit)

            assert resource.elapsed_time == self.ELAPSED_TIME
            assert resource.market_id == market_profit['marketId']
            assert resource.commission_applied == market_profit.get('commissionApplied')

            assert len(resource.profit_and_losses) == len(market_profit['profitAndLosses'])
            # todo complete

    def test_limit_order(self):
        limit_order = LimitOrder(1.01, 12, persistenceType='LIMIT', timeInForce=True, minFillSize=2,
                                 betTargetType='BACKERS_PROFIT', betTargetSize=3)
        assert limit_order.price == 1.01
        assert limit_order.size == 12
        assert limit_order.persistence_type == 'LIMIT'
        assert limit_order.time_in_force is True
        assert limit_order.min_fill_size == 2
        assert limit_order.bet_target_type == 'BACKERS_PROFIT'
        assert limit_order.bet_target_size == 3

    def test_limit_on_close_order(self):
        limit_on_close_order = LimitOnCloseOrder(liability=12, price=100)
        assert limit_on_close_order.liability == 12
        assert limit_on_close_order.price == 100

    def test_market_on_close_order(self):
        market_on_close_order = MarketOnCloseOrder(liability=12)
        assert market_on_close_order.liability == 12

    def test_place_orders(self):
        mock_response = create_mock_json('tests/resources/place_orders.json')
        place_orders = mock_response.json().get('result')
        resource = resources.PlaceOrders(elapsed_time=self.ELAPSED_TIME,
                                        **place_orders)
        assert resource.elapsed_time == self.ELAPSED_TIME
        assert resource.market_id == place_orders['marketId']
        assert resource.status == place_orders['status']
        assert resource.customer_ref == place_orders.get('customerRef')
        assert resource.error_code == place_orders.get('errorCode')
        assert len(resource.place_instruction_reports) == len(place_orders.get('instructionReports'))

        for order in place_orders.get('instructionReports'):
            assert resource.place_instruction_reports[0].size_matched == order['sizeMatched']
            assert resource.place_instruction_reports[0].status == order['status']
            assert resource.place_instruction_reports[0].bet_id == order['betId']
            assert resource.place_instruction_reports[0].average_price_matched == order['averagePriceMatched']
            assert resource.place_instruction_reports[0].placed_date == datetime.datetime.strptime(
                        order['placedDate'], "%Y-%m-%dT%H:%M:%S.%fZ")
            assert resource.place_instruction_reports[0].error_code == order.get('errorCode')

            assert resource.place_instruction_reports[0].instruction.selection_id == order['instruction']['selectionId']
            assert resource.place_instruction_reports[0].instruction.side == order['instruction']['side']
            assert resource.place_instruction_reports[0].instruction.order_type == order['instruction']['orderType']
            assert resource.place_instruction_reports[0].instruction.handicap == order['instruction']['handicap']

            assert resource.place_instruction_reports[0].instruction.limit_order.persistence_type == order['instruction']['limitOrder']['persistenceType']
            assert resource.place_instruction_reports[0].instruction.limit_order.price == order['instruction']['limitOrder']['price']
            assert resource.place_instruction_reports[0].instruction.limit_order.size == order['instruction']['limitOrder']['size']

    def test_cancel_orders(self):
        mock_response = create_mock_json('tests/resources/cancel_orders.json')
        cancel_orders = mock_response.json().get('result')
        resource = resources.CancelOrders(elapsed_time=self.ELAPSED_TIME,
                                          **cancel_orders)
        assert resource.elapsed_time == self.ELAPSED_TIME
        assert resource.market_id == cancel_orders['marketId']
        assert resource.status == cancel_orders['status']
        assert resource.customer_ref == cancel_orders.get('customerRef')
        assert resource.error_code == cancel_orders.get('errorCode')
        assert len(resource.cancel_instruction_reports) == len(cancel_orders.get('instructionReports'))

        for order in cancel_orders.get('instructionReports'):
            assert resource.cancel_instruction_reports[0].size_cancelled == order['sizeCancelled']
            assert resource.cancel_instruction_reports[0].status == order['status']
            assert resource.cancel_instruction_reports[0].cancelled_date == datetime.datetime.strptime(
                        order['cancelledDate'], "%Y-%m-%dT%H:%M:%S.%fZ")

            assert resource.cancel_instruction_reports[0].instruction.bet_id == order['instruction']['betId']
            assert resource.cancel_instruction_reports[0].instruction.size_reduction == order['instruction'].get('sizeReduction')

    def test_update_orders(self):
        mock_response = create_mock_json('tests/resources/update_orders.json')
        update_orders = mock_response.json().get('result')
        resource = resources.UpdateOrders(elapsed_time=self.ELAPSED_TIME,
                                          **update_orders)
        assert resource.elapsed_time == self.ELAPSED_TIME
        assert resource.market_id == update_orders['marketId']
        assert resource.status == update_orders['status']
        assert resource.customer_ref == update_orders.get('customerRef')
        assert resource.error_code == update_orders.get('errorCode')
        assert len(resource.update_instruction_reports) == len(update_orders.get('instructionReports'))

        for order in update_orders.get('instructionReports'):
            pass

    def test_replace_orders(self):
        mock_response = create_mock_json('tests/resources/replace_orders.json')
        replace_orders = mock_response.json().get('result')
        resource = resources.ReplaceOrders(elapsed_time=self.ELAPSED_TIME,
                                          **replace_orders)
        assert resource.elapsed_time == self.ELAPSED_TIME
        assert resource.market_id == replace_orders['marketId']
        assert resource.status == replace_orders['status']
        assert resource.customer_ref == replace_orders.get('customerRef')
        assert resource.error_code == replace_orders.get('errorCode')
        # assert len(resource.instruction_reports) == len(replace_orders.get('instructionReports'))



from .exceptions import StatusCodeError

TICK_SIZES = {1.0: 0.01, 2.0: 0.02, 3.0: 0.05, 4.0: 0.1, 6.0: 0.2, 10.0: 0.5,
              20.0: 1.0, 30.0: 2.0, 50.0: 5.0, 100.0: 10.0, 1000.0: 1000}


def check_status_code(response, codes=None):
    """
    Checks response.status_code is in codes.

    :param requests.request response: Requests response
    :param list codes: List of accepted codes or callable
    :raises: StatusCodeError if code invalid
    """
    codes = codes or [200]
    if response.status_code not in codes:
        raise StatusCodeError(response.status_code)


def clean_locals(data):
    """
    Clean up locals dict, remove empty and self/session/params params
    and convert to camelCase.

    :param {} data: locals dicts from a function.
    :returns: dict
    """
    if data.get('params') is not None:
        return data.get('params')
    else:
        return {
            to_camel_case(k): v for k, v in data.items() if v is not None and k not in
            ['self', 'session', 'params', 'lightweight']
        }


def to_camel_case(snake_str):
    """
    Converts snake_string to camelCase

    :param str snake_str:
    :returns: str
    """
    components = snake_str.split('_')
    return components[0] + "".join(x.title() for x in components[1:])



import threading
import socket
import ssl
import datetime
import collections

from ..exceptions import (
    SocketError,
    ListenerError,
)
from ..compat import (
    is_py3,
    json,
)


class BetfairStream(object):
    """Socket holder, connects to betfair and
    pushes any received data to listener
    """

    __port = 443
    __CRLF = '\r\n'
    __encoding = 'utf-8'

    HOSTS = collections.defaultdict(
        lambda: 'stream-api.betfair.com',
        integration='stream-api-integration.betfair.com',
    )

    def __init__(self, unique_id, listener, app_key, session_token, timeout, buffer_size, description, host):
        self._unique_id = unique_id
        self.listener = listener
        self.app_key = app_key
        self.session_token = session_token
        self.timeout = timeout
        self.buffer_size = buffer_size
        self.description = description
        self.host = self.HOSTS[host]
        self.receive_count = 0
        self.datetime_last_received = None

        self._socket = None
        self._running = False

    def start(self, async_=False):
        """Starts read loop, new thread if async and
        connects/authenticates if not already running.

        :param async_: If True new thread is started
        """
        if not self._running:
            self._connect()
            self.authenticate()
        if async_:
            t = threading.Thread(name=self.description, target=self._read_loop)
            t.daemon = False
            t.start()
        else:
            self._read_loop()

    def stop(self):
        """Stops read loop and closes socket if it has been created.
        """
        self._running = False

        if self._socket is None:
            return
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
        except socket.error:
            pass
        self._socket = None

    def authenticate(self):
        """Authentication request.
        """
        unique_id = self.new_unique_id()
        message = {
            'op': 'authentication',
            'id': unique_id,
            'appKey': self.app_key,
            'session': self.session_token,
        }
        self._send(message)
        return unique_id

    def heartbeat(self):
        """Heartbeat request to keep session alive.
        """
        unique_id = self.new_unique_id()
        message = {
            'op': 'heartbeat',
            'id': unique_id,
        }
        self._send(message)
        return unique_id

    def subscribe_to_markets(self, market_filter, market_data_filter, initial_clk=None, clk=None,
                             conflate_ms=None, heartbeat_ms=None, segmentation_enabled=True):
        """
        Market subscription request.

        :param dict market_filter: Market filter
        :param dict market_data_filter: Market data filter
        :param str initial_clk: Sequence token for reconnect
        :param str clk: Sequence token for reconnect
        :param int conflate_ms: conflation rate (bounds are 0 to 120000)
        :param int heartbeat_ms: heartbeat rate (500 to 5000)
        :param bool segmentation_enabled: allow the server to send large sets of data
        in segments, instead of a single block
        """
        unique_id = self.new_unique_id()
        message = {
            'op': 'marketSubscription',
            'id': unique_id,
            'marketFilter': market_filter,
            'marketDataFilter': market_data_filter,
            'initialClk': initial_clk,
            'clk': clk,
            'conflateMs': conflate_ms,
            'heartbeatMs': heartbeat_ms,
            'segmentationEnabled': segmentation_enabled,
        }
        if initial_clk and clk:
            # if resubscribe only update unique_id
            self.listener.stream_unique_id = unique_id
        else:
            self.listener.register_stream(unique_id, 'marketSubscription')
        self._send(message)
        return unique_id

    def subscribe_to_orders(self, order_filter=None, initial_clk=None, clk=None, conflate_ms=None,
                            heartbeat_ms=None, segmentation_enabled=True):
        """
        Order subscription request.

        :param dict order_filter: Order filter to be applied
        :param str initial_clk: Sequence token for reconnect
        :param str clk: Sequence token for reconnect
        :param int conflate_ms: conflation rate (bounds are 0 to 120000)
        :param int heartbeat_ms: heartbeat rate (500 to 5000)
        :param bool segmentation_enabled: allow the server to send large sets of data
        in segments, instead of a single block
        """
        unique_id = self.new_unique_id()
        message = {
            'op': 'orderSubscription',
            'id': unique_id,
            'orderFilter': order_filter,
            'initialClk': initial_clk,
            'clk': clk,
            'conflateMs': conflate_ms,
            'heartbeatMs': heartbeat_ms,
            'segmentationEnabled': segmentation_enabled,
        }
        if initial_clk and clk:
            # if resubscribe only update unique_id
            self.listener.stream_unique_id = unique_id
        else:
            self.listener.register_stream(unique_id, 'orderSubscription')
        self._send(message)
        return unique_id

    def new_unique_id(self):
        self._unique_id += 1
        return self._unique_id

    def _connect(self):
        """Creates socket and sets running to True.
        """
        self._socket = self._create_socket()
        self._running = True

    def _create_socket(self):
        """Creates ssl socket, connects to stream api and
        sets timeout.
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s = ssl.wrap_socket(s)
        s.connect((self.host, self.__port))
        s.settimeout(self.timeout)
        return s

    def _read_loop(self):
        """Read loop, splits by CRLF and pushes received data
        to _data.
        """
        while self._running:
            received_data_raw = self._receive_all()
            if self._running:
                self.receive_count += 1
                self.datetime_last_received = datetime.datetime.utcnow()
                received_data_split = received_data_raw.split(self.__CRLF)
                for received_data in received_data_split:
                    if received_data:
                        self._data(received_data)

    def _receive_all(self):
        """Whilst socket is running receives data from socket,
        till CRLF is detected.
        """
        (data, part) = ('', '')
        if is_py3:
            crlf_bytes = bytes(self.__CRLF, encoding=self.__encoding)
        else:
            crlf_bytes = self.__CRLF

        while self._running and part[-2:] != crlf_bytes:
            try:
                part = self._socket.recv(self.buffer_size)
            except (socket.timeout, socket.error) as e:
                if self._running:
                    self.stop()
                    raise SocketError('[Connect: %s]: Socket %s' % (self._unique_id, e))
                else:
                    return  # 133, prevents error if stop is called mid recv

            # an empty string indicates the server shutdown the socket
            if len(part) == 0:
                self.stop()
                raise SocketError('Connection closed by server')

            data += part.decode(self.__encoding)
        return data

    def _data(self, received_data):
        """Sends data to listener, if False is returned; socket
        is closed.

        :param received_data: Decoded data received from socket.
        """
        if self.listener.on_data(received_data) is False:
            self.stop()
            raise ListenerError(self.listener.connection_id, received_data)

    def _send(self, message):
        """If not running connects socket and
        authenticates. Adds CRLF and sends message
        to Betfair.

        :param message: Data to be sent to Betfair.
        """
        if not self._running:
            self._connect()
            self.authenticate()
        message_dumped = json.dumps(message) + self.__CRLF
        try:
            self._socket.send(message_dumped.encode())
        except (socket.timeout, socket.error) as e:
            self.stop()
            raise SocketError('[Connect: %s]: Socket %s' % (self._unique_id, e))

    def __str__(self):
        return '<BetfairStream [%s]>' % ('running' if self._running else 'not running')

    def __repr__(self):
        return '<BetfairStream>'


class HistoricalStream(object):
    """Copy of 'Betfair Stream' for parsing
    historical data.
    """

    def __init__(self, directory, listener):
        """
        :param str directory: Directory of betfair data
        :param BaseListener listener: Listener object
        """
        self.directory = directory
        self.listener = listener
        self._running = False

    def start(self, async_=False):
        self._running = True
        if async_:
            t = threading.Thread(name='HistoricalStream', target=self._read_loop)
            t.daemon = False
            t.start()
        else:
            self._read_loop()

    def stop(self):
        self._running = False

    def _read_loop(self):
        with open(self.directory, 'r') as f:
            for update in f:
                if self.listener.on_data(update) is False:
                    # if on_data returns an error stop the stream and raise error
                    self.stop()
                    raise ListenerError('HISTORICAL', update)
                if not self._running:
                    break
            else:
                # if f has finished, also stop the stream
                self.stop()



import unittest
from betfairlightweight import metadata


class MetadataTest(unittest.TestCase):

    def test_transaction_limit(self):
        assert metadata.transaction_limit == 1000

    def test_limits(self):
        assert metadata.order_limits == {
            'placeOrders': 200,
            'cancelOrders': 60,
            'updateOrders': 60,
            'replaceOrders': 60
        }
        assert metadata.list_current_orders == {
            'marketIds': 250,
            'orders': 1000
        }
        assert metadata.list_market_catalogue == {
            'MARKET_DESCRIPTION': 1,
            'RUNNER_DESCRIPTION': 0,
            'EVENT': 0,
            'EVENT_TYPE': 0,
            'COMPETITION':	0,
            'RUNNER_METADATA': 1,
            'MARKET_START_TIME': 0
        }

        assert metadata.list_market_book == {
            '': 2,
            'SP_AVAILABLE': 3,
            'EX_BEST_OFFERS': 5,
            'SP_TRADED': 7,
            'EX_ALL_OFFERS': 17,
            'EX_TRADED': 17
        }

        assert metadata.list_market_profit_and_loss == {
            '': 4
        }

    def test_currency_parameters(self):
        assert metadata.currency_parameters == {
            'GBP': {
                'min_bet_size': 2,
                'min_bet_payout': 10,
                'min_bsp_liability': 10
            },
            'EUR': {
                'min_bet_size': 2,
                'min_bet_payout': 20,
                'min_bsp_liability': 20
            },
            'USD': {
                'min_bet_size': 4,
                'min_bet_payout': 20,
                'min_bsp_liability': 20
            },
            'HKD': {
                'min_bet_size': 25,
                'min_bet_payout': 125,
                'min_bsp_liability': 125
            },
            'AUD': {
                'min_bet_size': 5,
                'min_bet_payout': 30,
                'min_bsp_liability': 30
            },
            'CAD': {
                'min_bet_size': 6,
                'min_bet_payout': 30,
                'min_bsp_liability': 30
            },
            'DKK': {
                'min_bet_size': 30,
                'min_bet_payout': 150,
                'min_bsp_liability': 150
            },
            'NOK': {
                'min_bet_size': 30,
                'min_bet_payout': 150,
                'min_bsp_liability': 150
            },
            'SEK': {
                'min_bet_size': 30,
                'min_bet_payout': 150,
                'min_bsp_liability': 150
            },
            'SGD': {
                'min_bet_size': 6,
                'min_bet_payout': 30,
                'min_bsp_liability': 30
            },
            'RON': {
                'min_bet_size': 10,
                'min_bet_payout': 50,
                'min_bsp_liability': (10, 50)  # (back, lay)
            },
        }



import datetime
import ujson as json
import unittest

from betfairlightweight.resources.baseresource import BaseResource
from tests.unit.tools import create_mock_json


class BaseResourceInit(unittest.TestCase):

    def test_init(self):
        base_resource = BaseResource()
        assert base_resource._datetime_created is not None
        assert base_resource._datetime_updated is not None
        assert base_resource.elapsed_time is None
        assert base_resource._data == {}

    def test_data(self):
        mock_response = create_mock_json('tests/resources/base_resource.json')
        base_resource = BaseResource(elapsed_time=1.2,
                                     **mock_response.json())

        assert base_resource.elapsed_time == 1.2
        assert base_resource._data == mock_response.json()

    def test_data_json(self):
        mock_response = create_mock_json('tests/resources/base_resource.json')
        base_resource = BaseResource(elapsed_time=1.2,
                                     **mock_response.json())
        assert base_resource.json() == json.dumps(mock_response.json())

    def test_strip_datetime(self):
        base_resource = BaseResource()
        for string in ['2100-06-01T10:10:00.000Z', '2100-06-01T10:10:00.00Z', '2100-06-01T10:10:00.0Z']:
            stripped = base_resource.strip_datetime(string)
            assert type(stripped) == datetime.datetime

        stripped = base_resource.strip_datetime(None)
        assert not stripped

        stripped = base_resource.strip_datetime('45')
        assert not stripped

        integer = 1465631675000
        stripped = base_resource.strip_datetime(integer)
        assert type(stripped) == datetime.datetime

        stripped = base_resource.strip_datetime(None)
        assert not stripped

        stripped = base_resource.strip_datetime('45')
        assert not stripped

        # stripped = base_resource.strip_datetime(-1230000000345446)
        # assert not stripped  py3.6 is able to strip this

    def test_str_repr(self):
        base_resource = BaseResource()
        assert str(base_resource) == 'BaseResource'
        assert repr(base_resource) == '<BaseResource>'



import unittest

from betfairlightweight.streaming.stream import BaseStream, MarketStream, OrderStream
from tests import mock
from tests.unit.tools import create_mock_json


class BaseStreamTest(unittest.TestCase):

    def setUp(self):
        self.listener = mock.Mock()
        self.listener.max_latency = 0.5
        self.stream = BaseStream(self.listener)

    def test_init(self):
        assert self.stream._listener == self.listener

        assert self.stream._initial_clk is None
        assert self.stream._clk is None
        assert self.stream._caches == {}
        assert self.stream._updates_processed == 0
        assert self.stream.time_created is not None
        assert self.stream.time_updated is not None

    @mock.patch('betfairlightweight.streaming.stream.BaseStream._process')
    @mock.patch('betfairlightweight.streaming.stream.BaseStream._update_clk')
    def test_on_subscribe(self, mock_update_clk, mock_process):
        self.stream.on_subscribe({})
        mock_update_clk.assert_called_once_with({})

        self.stream.on_subscribe({'mc': {123}})
        mock_process.assert_called_once_with({123}, None)

    @mock.patch('betfairlightweight.streaming.stream.BaseStream._update_clk')
    def test_on_heartbeat(self, mock_update_clk):
        self.stream.on_heartbeat({})
        mock_update_clk.assert_called_once_with({})

    @mock.patch('betfairlightweight.streaming.stream.BaseStream.on_update')
    def test_on_resubscribe(self, mock_on_update):
        self.stream.on_resubscribe({})
        mock_on_update.assert_called_once_with({})

    @mock.patch('betfairlightweight.streaming.stream.BaseStream._process')
    @mock.patch('betfairlightweight.streaming.stream.BaseStream._calc_latency', return_value=0.1)
    @mock.patch('betfairlightweight.streaming.stream.BaseStream._update_clk')
    def test_on_update(self, mock_update_clk, mock_calc_latency, mock_process):
        mock_response = create_mock_json('tests/resources/streaming_mcm_update.json')
        self.stream.on_update(mock_response.json())

        mock_update_clk.assert_called_with(mock_response.json())
        mock_calc_latency.assert_called_with(mock_response.json().get('pt'))
        mock_process.assert_called_with(mock_response.json().get('mc'), mock_response.json().get('pt'))

        mock_calc_latency.return_value = 10
        self.stream.on_update(mock_response.json())

    def test_clear_cache(self):
        self.stream._caches = {1: 'abc'}
        self.stream.clear_cache()

        assert self.stream._caches == {}

    def test_snap(self):
        market_books = self.stream.snap()
        assert market_books == []

        mock_cache = mock.Mock()
        mock_cache.market_id = '1.1'
        self.stream._caches = {'1.1': mock_cache}

        market_books = self.stream.snap()
        assert market_books == [mock_cache.create_resource()]

        market_books = self.stream.snap(['1.2'])
        assert market_books == []

        market_books = self.stream.snap(['1.1'])
        assert market_books == [mock_cache.create_resource()]

    def test_snap_dict_size_err(self):
        mock_cache = mock.Mock()
        mock_cache.market_id = '1.1'

        def _change_dict(*_, **__):
            self.stream._caches['1.{}'.format(len(self.stream._caches))] = \
                mock_cache

        mock_cache.create_resource = _change_dict
        self.stream._caches = {'1.{}'.format(i): mock_cache for i in range(2)}

        self.stream.snap()

    def test_on_creation(self):
        self.stream._on_creation()

    def test_process(self):
        self.stream._process(None, None)

    def test_on_process(self):
        self.stream.on_process([1, 2])
        self.stream.output_queue.put.assert_called_with([1, 2])

    def test_update_clk(self):
        self.stream._update_clk({'initialClk': 1234})
        assert self.stream._initial_clk == 1234

        self.stream._update_clk({'clk': 123})
        assert self.stream._clk == 123

    def test_unique_id(self):
        assert self.stream.unique_id == self.listener.stream_unique_id

    def test_output_queue(self):
        assert self.stream.output_queue == self.listener.output_queue

    def test_max_latency(self):
        assert self.stream._max_latency == self.listener.max_latency

    def test_lightweight(self):
        assert self.stream._lightweight == self.listener.lightweight

    @mock.patch('time.time', return_value=1485554805.107185)
    def test_calc_latency(self, mock_time):
        pt = 1485554796455
        assert self.stream._calc_latency(pt) is not None
        assert abs(self.stream._calc_latency(pt) - 8.652184) < 1e-5

    def test_len(self):
        assert len(self.stream) == 0

    def test_str(self):
        assert str(self.stream) == 'BaseStream'

    def test_repr(self):
        assert repr(self.stream) == '<BaseStream>'


class MarketStreamTest(unittest.TestCase):

    def setUp(self):
        self.listener = mock.Mock()
        self.stream = MarketStream(self.listener)

    @mock.patch('betfairlightweight.streaming.stream.MarketStream._process')
    @mock.patch('betfairlightweight.streaming.stream.MarketStream._update_clk')
    def test_on_subscribe(self, mock_update_clk, mock_process):
        self.stream.on_subscribe({})
        mock_update_clk.assert_called_once_with({})

        self.stream.on_subscribe({'mc': {123}})
        mock_process.assert_called_once_with({123}, None)

    @mock.patch('betfairlightweight.streaming.stream.MarketStream._process')
    @mock.patch('betfairlightweight.streaming.stream.MarketStream._update_clk')
    def test_on_subscribe(self, mock_update_clk, mock_process):
        self.stream.on_subscribe({})
        mock_update_clk.assert_called_once_with({})

        self.stream.on_subscribe({'mc': {123}})
        mock_process.assert_called_once_with({123}, None)

    # @mock.patch('betfairlightweight.streaming.stream.MarketBookCache')
    # def test_process(self, mock_market_book_cache):
    #     now = mock.Mock()
    #     market_books = [mock.Mock()]
    #     self.stream._caches = mock.Mock()
    #     # self.stream._process(market_books, now)

    def test_str(self):
        assert str(self.stream) == 'MarketStream'

    def test_repr(self):
        assert repr(self.stream) == '<MarketStream [0]>'


class OrderStreamTest(unittest.TestCase):

    def setUp(self):
        self.listener = mock.Mock()
        self.stream = OrderStream(self.listener)

    @mock.patch('betfairlightweight.streaming.stream.OrderStream._process')
    @mock.patch('betfairlightweight.streaming.stream.OrderStream._update_clk')
    def test_on_subscribe(self, mock_update_clk, mock_process):
        self.stream.on_subscribe({})
        mock_update_clk.assert_called_once_with({})

        self.stream.on_subscribe({'oc': {123}})
        mock_process.assert_called_once_with({123}, None)

    # def test_process(self):
    #     pass

    def test_str(self):
        assert str(self.stream) == 'OrderStream'

    def test_repr(self):
        assert repr(self.stream) == '<OrderStream [0]>'



import requests
import collections
import datetime
import os

from .exceptions import (
    PasswordError,
    AppKeyError,
    CertsError,
)
from .compat import FileNotFoundError


class BaseClient(object):
    """
    Base API client
    """

    IDENTITY_URLS = collections.defaultdict(
        lambda: 'https://identitysso.betfair.com/api/',
        spain='https://identitysso.betfair.es/api/',
        italy='https://identitysso.betfair.it/api/',
        romania='https://identitysso.betfair.ro/api/',
        sweden='https://identitysso.betfair.se/api/',
        australia='https://identitysso.betfair.au/api/',
    )

    IDENTITY_CERT_URLS = collections.defaultdict(
        lambda: 'https://identitysso-cert.betfair.com/api/',
        spain='https://identitysso-cert.betfair.es/api/',
        italy='https://identitysso-cert.betfair.it/api/',
        romania='https://identitysso-cert.betfair.ro/api/',
        sweden='https://identitysso-cert.betfair.se/api/',
    )

    API_URLS = collections.defaultdict(
        lambda: 'https://api.betfair.com/exchange/'
    )

    NAVIGATION_URLS = collections.defaultdict(
        lambda: 'https://api.betfair.com/exchange/betting/rest/v1/en/navigation/menu.json',
        spain='https://api.betfair.es/exchange/betting/rest/v1/en/navigation/menu.json',
        italy='https://api.betfair.it/exchange/betting/rest/v1/en/navigation/menu.json'
    )

    def __init__(self, username, password=None, app_key=None, certs=None, locale=None, cert_files=None,
                 lightweight=False):
        """
        Creates base client for API operations.

        :param str username: Betfair username
        :param str password: Password for supplied username, if None will look in .bashprofile
        :param str app_key: App Key for account, if None will look in .bashprofile
        :param str certs: Directory for certificates, if None will look in /certs/
        :param str locale: Exchange to be used, defaults to UK for login and global for api
        :param list cert_files: Certificate and key files. If None will look in `certs`
        :param bool lightweight: If True endpoints will return dict not a resource (22x faster)
        """
        self.username = username
        self.password = password
        self.app_key = app_key
        self.certs = certs
        self.locale = locale
        self.cert_files = cert_files
        self.lightweight = lightweight

        self.session = requests
        self._login_time = None
        self.session_token = None
        self.identity_uri = self.IDENTITY_URLS[locale]
        self.identity_cert_uri = self.IDENTITY_CERT_URLS[locale]
        self.api_uri = self.API_URLS[locale]
        self.navigation_uri = self.NAVIGATION_URLS[locale]

        self.get_password()
        self.get_app_key()

    def set_session_token(self, session_token):
        """
        Sets session token and new login time.

        :param str session_token: Session token from request.
        """
        self.session_token = session_token
        self._login_time = datetime.datetime.now()

    def get_password(self):
        """
        If password is not provided will look in environment variables
        for username+'password'.
        """
        if self.password is None:
            if os.environ.get(self.username+'password'):
                self.password = os.environ.get(self.username+'password')
            else:
                raise PasswordError(self.username)

    def get_app_key(self):
        """
        If app_key is not provided will look in environment
        variables for username.
        """
        if self.app_key is None:
            if os.environ.get(self.username):
                self.app_key = os.environ.get(self.username)
            else:
                raise AppKeyError(self.username)

    def client_logout(self):
        """
        Resets session token and login time.
        """
        self.session_token = None
        self._login_time = None

    @property
    def session_expired(self):
        """
        Returns True if login_time not set or seconds since
        login time is greater than 200 mins.
        """
        if not self._login_time or (datetime.datetime.now()-self._login_time).total_seconds() > 12000:
            return True

    @property
    def cert(self):
        """
        The betfair certificates, by default it looks for the
        certificates in /certs/.

        :return: Path of cert files
        :rtype: str
        """
        if self.cert_files is not None:
            return self.cert_files

        certs = self.certs or '/certs/'
        ssl_path = os.path.join(os.pardir, certs)
        try:
            cert_path = os.listdir(ssl_path)
        except FileNotFoundError:
            raise CertsError(certs)
        except OSError:   # Python 2 compatability
            raise CertsError(certs)

        cert = None
        key = None
        for file in cert_path:
            ext = os.path.splitext(file)[-1]
            if ext in ['.crt', '.cert']:
                cert = os.path.join(ssl_path, file)
            elif ext == '.key':
                key = os.path.join(ssl_path, file)
        if cert is None or key is None:
            raise CertsError(certs)
        return [cert, key]

    @property
    def login_headers(self):
        return {
            'Accept': 'application/json',
            'X-Application': self.app_key,
            'content-type': 'application/x-www-form-urlencoded'
        }

    @property
    def keep_alive_headers(self):
        return {
            'Accept': 'application/json',
            'X-Application': self.app_key,
            'X-Authentication': self.session_token,
            'content-type': 'application/x-www-form-urlencoded'
        }

    @property
    def request_headers(self):
        return {
            'X-Application': self.app_key,
            'X-Authentication': self.session_token,
            'content-type': 'application/json',
            'Accept-Encoding': 'gzip, deflate',
            'Connection': 'keep-alive',
            'User-Agent': 'betfairlightweight',
        }



import unittest
from tests import mock
from requests.exceptions import ConnectionError

from betfairlightweight import APIClient
from betfairlightweight.endpoints.inplayservice import InPlayService
from betfairlightweight.exceptions import APIError, InvalidResponse


class InPlayServiceTest(unittest.TestCase):

    def setUp(self):
        self.client = APIClient('username', 'password', 'app_key', 'UK')
        self.in_play_service = InPlayService(self.client)

    def test_init(self):
        assert self.in_play_service.connect_timeout == 3.05
        assert self.in_play_service.read_timeout == 16
        assert self.in_play_service._error == APIError
        assert self.in_play_service.client == self.client

    @mock.patch('betfairlightweight.endpoints.inplayservice.InPlayService.process_response')
    @mock.patch('betfairlightweight.endpoints.inplayservice.InPlayService.request', return_value=(mock.Mock(), 1.3))
    def test_get_event_timeline(self, mock_request, mock_process_response):
        event_id = 12345
        params = {
            'eventId': event_id,
            'alt': 'json',
            'regionCode': 'UK',
            'locale': 'en_GB'
        }
        self.in_play_service.get_event_timeline(event_id)

        mock_request.assert_called_with(
                url='https://www.betfair.com/inplayservice/v1.1/eventTimeline', session=None, params=params)
        assert mock_request.call_count == 1
        assert mock_process_response.call_count == 1

    @mock.patch('betfairlightweight.endpoints.inplayservice.InPlayService.process_response')
    @mock.patch('betfairlightweight.endpoints.inplayservice.InPlayService.request', return_value=(mock.Mock(), 1.3))
    def test_get_event_timelines(self, mock_request, mock_process_response):
        event_ids = [12345, 54321]
        params = {
            'eventIds': '12345,54321',
            'alt': 'json',
            'regionCode': 'UK',
            'locale': 'en_GB'
        }
        self.in_play_service.get_event_timelines(event_ids)

        mock_request.assert_called_with(
                url='https://www.betfair.com/inplayservice/v1.1/eventTimelines', session=None, params=params)
        assert mock_request.call_count == 1
        assert mock_process_response.call_count == 1

    @mock.patch('betfairlightweight.endpoints.inplayservice.InPlayService.process_response')
    @mock.patch('betfairlightweight.endpoints.inplayservice.InPlayService.request', return_value=(mock.Mock(), 1.3))
    def test_get_scores(self, mock_request, mock_process_response):
        event_ids = [12345]
        params = {
            'eventIds': ','.join(str(x) for x in event_ids),
            'alt': 'json',
            'regionCode': 'UK',
            'locale': 'en_GB'
        }
        self.in_play_service.get_scores(event_ids)

        mock_request.assert_called_with(
                url='https://www.betfair.com/inplayservice/v1.1/scores', session=None, params=params)
        assert mock_request.call_count == 1
        assert mock_process_response.call_count == 1

    @mock.patch('betfairlightweight.endpoints.inplayservice.check_status_code')
    @mock.patch('betfairlightweight.endpoints.inplayservice.InPlayService.headers')
    @mock.patch('betfairlightweight.baseclient.requests.get')
    def test_request(self, mock_get, mock_headers, mock_check_status_code):
        params = [1, 2, 3]
        url = '123'

        self.in_play_service.request(params=params, url=url)

        mock_get.assert_called_with(
                url, headers=mock_headers, params=params)
        assert mock_get.call_count == 1
        assert mock_check_status_code.call_count == 1

    @mock.patch('betfairlightweight.endpoints.inplayservice.InPlayService.headers')
    @mock.patch('betfairlightweight.baseclient.requests.get')
    def test_request_error(self, mock_get, mock_headers):
        params = [1, 2, 3]
        url = '123'
        mock_get.side_effect = ConnectionError()
        with self.assertRaises(APIError):
            self.in_play_service.request(params=params, url=url)

        mock_get.side_effect = ValueError()
        with self.assertRaises(APIError):
            self.in_play_service.request(params=params, url=url)

    @mock.patch('betfairlightweight.endpoints.inplayservice.check_status_code')
    @mock.patch('betfairlightweight.endpoints.inplayservice.InPlayService.headers')
    @mock.patch('betfairlightweight.baseclient.requests.get')
    def test_request_json_error(self, mock_get, mock_headers, mock_check_status_code):
        params = [1, 2, 3]
        url = '123'

        response = mock.Mock()
        mock_get.return_value = response
        response.json.side_effect = ValueError()

        with self.assertRaises(InvalidResponse):
            self.in_play_service.request(params=params, url=url)

    def test_headers(self):
        assert self.in_play_service.headers == {
            'Connection': 'keep-alive',
            'Content-Type': 'application/json'
        }

    def test_url(self):
        assert self.in_play_service.url == 'https://www.betfair.com/inplayservice/v1.1/'



from mock import Mock
import json


def create_mock_json(path, status_code=200):
    with open(path) as f:
        resp = Mock()
        resp.content = f.read()
        resp.json.return_value = json.loads(resp.content)
        resp.status_code = status_code
        return resp
